//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "TraverseParallel.h"
#include "OpenMP.h"
#include "Mesh.h"

#ifdef _OPENMP

namespace AMDiS {

  TraverseParallelStack::TraverseParallelStack(int n, int mode)
    : nThreads(0),
      parallelMode(mode)
  {
    if (n == 0)
      nThreads = omp_get_overall_max_threads();
    else
      nThreads = n;
    
    stacks.resize(nThreads);

    if (parallelMode == 0) {
      int i = 0;
      for (std::vector<TraverseStack*>::iterator it = stacks.begin();
	   it != stacks.end(); ++it) {
	
	(*it) = new TraverseStack();
	(*it)->setMyThreadId(i++);
	(*it)->setMaxThreads(nThreads);
      }
    } else {
      for (std::vector<TraverseStack*>::iterator it = stacks.begin();
	   it != stacks.end(); ++it)	
	(*it) = new TraverseStack();      
    }
  }

  TraverseParallelStack::~TraverseParallelStack()
  {
    for (std::vector<TraverseStack*>::iterator it = stacks.begin();
	 it != stacks.end(); ++it)
      delete (*it);
  }

  ElInfo* TraverseParallelStack::traverseFirst(Mesh *mesh, int level, Flag fill_flag)
  {
    FUNCNAME("TraverseParallelStack::traverseFirst()");

    TEST_EXIT(fill_flag.isSet(Mesh::CALL_LEAF_EL))("not yet implemented");
    TEST_EXIT(stacks[omp_get_thread_num()])
      ("something wrong with parallel stack traverse");

    if (parallelMode == 0) {
      return stacks[omp_get_thread_num()]->traverseFirst(mesh, level, fill_flag);
    } else {
      ElInfo *elInfo = stacks[omp_get_thread_num()]->traverseFirst(mesh, level, fill_flag);
      while (elInfo &&
	     elInfo->getElement()->getIndex() % nThreads != omp_get_thread_num()) {
	elInfo = stacks[omp_get_thread_num()]->traverseNext(elInfo);
      }
      return elInfo;
    }
  }

  ElInfo* TraverseParallelStack::traverseNext(ElInfo* elInfoOld)
  {
    if (parallelMode == 0) {
      return stacks[omp_get_thread_num()]->traverseNext(elInfoOld);
    } else {
      ElInfo *elInfo = stacks[omp_get_thread_num()]->traverseNext(elInfoOld);
      while (elInfo &&
	     elInfo->getElement()->getIndex() % nThreads != omp_get_thread_num()) {
	elInfo = stacks[omp_get_thread_num()]->traverseNext(elInfo);
      }
      return elInfo;
    }      
  }

}

#endif // _OPENMP
